#include "dirichlet.h"

Dirichlet::Dirichlet(){}

Dirichlet::Dirichlet(int k, vector<double> alpha, double alpha_sum){

    assert(k >= 1 && k <= 3);
    assert(k == alpha.size());

    this->k = k;

    bool isValidConcparams = true;
    for (int i = 0; i < alpha.size(); i++) {
        if(alpha[i] < 0) {
            isValidConcparams = false;
            break;
        }
        alpha_lgammas.push_back(lgamma(alpha[i]));
    } 
    if(!isValidConcparams) return;

    this->alpha = alpha;
    this->alpha_sum = alpha_sum;

    init();
}

void Dirichlet::init(){
    assert(k > 0);
    assert(alpha_sum > 0);
    assert(alpha.size() == k);
    assert(k == 2 || k == 3);

    // part of 3rd term in det fisher of I(alpha)
    kappa_polygamma = polygamma(1, alpha_sum);

    // Equation 5.1 for Eq 5.4 for I(theta | alpha)
    kappa_lgamma = lgamma(alpha_sum);

    // 2nd term of I(alpha)
    k2log_lattice_const = ((double) k)/2.0 * log(lattice_constant(k));

    // Equation 5.4 2nd term of I(theta|alpha) 
    kover2 = ((double) k)/2.0;

    // h(mean) term of 1st term of I(alpha) 
    switch(k){
        case 2:
            I_mean = log(sqrt(2.0));
            break;
        case 3:  
            // second term actually belongs to I_kappa, but included here 
            I_mean = log(sqrt(3.0)/2.0) + log(M_PI/4.0); 
            break;
        default: 
            cerr << "Invalid number of dirichlet parameters : " << k << endl;
            exit(1);
    }

    // Equation 5.4 k-1 log(C_(k-1)) - lattice constant
    klog_lattice_const = ((double) (k-1)) * log(lattice_constant(k-1));

}

// Equation after 5.3
double Dirichlet::I_alpha(const int& n_bin_align){
    //I_prior = I_mean + I_kappa
    double I_kappa = (((double) (k+1))/2.0) * log(1.0 + pow(alpha_sum, 2.0)) - (((double) (k-1)) * log(alpha_sum));

    return  I_mean + I_kappa + k2log_lattice_const + (0.5 * log_det_fisher(n_bin_align));
}

// // Equation 5.2 - 3rd term of I(alpha)
double Dirichlet::log_det_fisher(const int& n_bin_align){
    assert(k > 0);

    double log_polygamma_alpha = 0.0;
    double reciprocal_polygamma_alpha = 0.0;
    assert(alpha.size() == k);

    // 1st term in det fisher of I(alpha)
    double klogN = ((double) k) * log(n_bin_align);

    for(int i = 0; i < k; i++){
        double polygamma_alpha_i = polygamma(1, alpha[i]);
        log_polygamma_alpha += log(polygamma_alpha_i);
        reciprocal_polygamma_alpha += 1.0 / polygamma_alpha_i;
    }

    return klogN + log_polygamma_alpha + log(1 - kappa_polygamma * reciprocal_polygamma_alpha);
}

// Equation 5.4 + Equation 5.5
double Dirichlet::I_theta_plus_I_A(SSEAlign& alignobj, const vector<int>& trans){
    assert(trans.size() == k);
    double sumtrans = accumulate(trans.begin(), trans.end(), 0.0);

    double det_fisher = (double) (k - 1) * log(sumtrans);

    double thetaSum = 0;
    vector<double> theta(k, 0);

    // Equation 5.1
    double alpha_theta = 0;
    double loggamma_alpha = 0;
    double log_theta_i;
    
    //Equation 5.5
    double I_a = 0;

    // Equation before 5.5 - theta computation from dirichlets
    for(int i = 0; i < (k - 1); i++){

        theta[i] = ((double) trans[i] + alpha[i] - 0.5) 
                 / ((double) sumtrans + alpha_sum - (double)k/2.0);
        thetaSum += theta[i];  
        log_theta_i = log(theta[i]);   
        det_fisher += -log_theta_i;   
        alpha_theta += (alpha[i] - 1.0) * log_theta_i;
        loggamma_alpha -= alpha_lgammas[i];
        I_a += (-log_theta_i * trans[i]);
    }
    theta[k-1] = 1.0 - thetaSum;
    log_theta_i = log(theta[k-1]);
    det_fisher += -log_theta_i;
    alpha_theta += (alpha[k-1] - 1.0) * log_theta_i;
    loggamma_alpha -= alpha_lgammas[k-1]; 
    I_a += (-log_theta_i * trans[k-1]);

    double log_pdf_sqrd = 2.0 * (kappa_lgamma + alpha_theta + loggamma_alpha);

    double det_over_log_pdf_sqrd = det_fisher + klog_lattice_const - log_pdf_sqrd;

    //Equation 5.4 + Equation 5.5
    return 0.5 * log1p(exp(det_over_log_pdf_sqrd)) + I_a + ((double) k - 1.0)/2.0;
}

// Equation 5.5 - I(A | theta, alpha)
double Dirichlet::I_A(SSEAlign& alignobj, const vector<int>& trans){
    assert(trans.size() == k);
    double sumtrans = accumulate(trans.begin(), trans.end(), 0.0);

    double thetaSum = 0;
    vector<double> theta(k, 0);

    double I_a = 0;
    double log_theta_i;

    // theta computation from dirichlets
    for(int i = 0; i < (k - 1); i++){

        theta[i] = ((double) trans[i] + alpha[i] - 0.5) 
                 / ((double) sumtrans + alpha_sum - (double)k/2.0);
        thetaSum += theta[i];  
        log_theta_i = log(theta[i]);   
        I_a += (-log_theta_i * trans[i]);
    }
    theta[k-1] = 1.0 - thetaSum;
    log_theta_i = log(theta[k-1]);
    I_a += (-log_theta_i * trans[k-1]);
    
    return I_a;
}

double Dirichlet::constTerms(const int& n_bin_align){
    // Equation 5.4 2nd term of I(theta|alpha) + Equation 5.5 last term of I(A)
    return kover2 + (double (n_bin_align * (k-1))) /2.0; 
}

vector<double> Dirichlet::getAlpha(){
    assert(alpha.size() == k);
    return alpha;
}

double Dirichlet::getKappa(){
    double sum = accumulate(alpha.begin(), alpha.end(), 0.0);
    assert(abs(sum - alpha_sum) < 1e-10);
    return alpha_sum;
}

vector<double> Dirichlet::getMean(){
    vector<double> meanvec;
    for(double al : alpha){
        meanvec.push_back(al/alpha_sum);
    }
    assert(meanvec.size() == k);
    return meanvec;
}

int Dirichlet::getK(){
    return k;
}

bool Dirichlet::isValid(){
    if(isEmpty()) return false;
    
    for(double alph : alpha){
        if(alph <= 1.0) return false;
    }
    return true;
}

string Dirichlet::print(){
    stringstream ss;
    int prec = 15;
    
    ss << "K         = " << k << endl;
    ss << "Alpha     = ";
    for (int i = 0;  i < alpha.size(); i++) {
        ss << setw(prec+4) << setprecision(prec) 
           << alpha[i] << "\t";
    } ss << endl;
    ss << "Alpha sum = " << setw(prec+4)  
       << setprecision(prec) <<alpha_sum << endl;

    return ss.str();
}

bool Dirichlet::isEmpty(){
    return (alpha.size() == 0);
}


// Equation between 5.4 & 5.5
vector<double> Dirichlet::computeThetaMode(){

    double thetaSum = 0;
    vector<double> theta(k, 0);

    assert(alpha.size() == k);
    for(int i = 0; i < k; i++){
        theta[i] = (alpha[i] - 1.0) 
                 / (alpha_sum - (double)k);
        thetaSum += theta[i];  
    }

    for(int i = 0; i < k; i++){
        theta[i] /= thetaSum;
    }
  
    return theta;
}

vector<double> Dirichlet::computeThetaMean(){

    double thetaSum = 0;
    vector<double> theta(k, 0);

    assert(alpha.size() == k);
    for(int i = 0; i < k; i++){
        theta[i] = alpha[i] / alpha_sum ;
        thetaSum += theta[i];  
    }

    for(int i = 0; i < k; i++){
        theta[i] /= thetaSum;
    }
  
    return theta;
}