#include "emMat.h"

EMMat::EMMat(){
    srand(0);
    random_number_generator.seed(0);
}

EMMat::EMMat(Submat& ssemat, DirManager& dirmg, const string alignfile, AAAlphabet& alph, int timeUB){
        
    srand(0);
    random_number_generator.seed(0);
   
    this->initmat = ssemat;
    this->ssealph = alph; 
    this->dir = dirmg;
    this->timeUB = timeUB;
    alignments = parseAlignmentData(alignfile, ssealph);
    assert(alignments.size() > 0);
    AAAlphabet submatAlph = initmat.getAlphabet();

    assert(submatAlph.size() == ssealph.size());
    assert(submatAlph.getAlphabet() == ssealph.getAlphabet()); 
}

pair<double, int> EMMat::cost_objective_function(Submat& ssemat){

    double algnCost = 0, dirCostM = 0, dirCostI = 0;
    double objectiveCost = 0;
    int inferredN;
    int isValid = 1;

    // I(M)
    double matCost = ssemat.compute_I_M();
    objectiveCost += matCost;
   
    //search for the optimal time for data
    for(int i = 0; i < alignments.size(); i++){

        //cost of matched regions in alignments (indel excluded [constant])
        // I(<S_i,T_j> | A_i, M, P, t_i)
        pair<double, int> searchres = searchForTotalProb(alignments[i], inferredN, ssemat, timeUB);
        algnCost = searchres.first;
        if(searchres.second == 0) {
            isValid = (int) searchres.second;
            break;
        }

        //match - I(theta_i | alpha_i, t_i) + I(A_i | theta_i, t_i)
        vector<int> transitions = alignments[i].getMTransitionVec();
        dirCostM = dir.getMatchDirichlet(inferredN).I_theta_plus_I_A(alignments[i], transitions);

        // indel - I(theta_i | alpha_i, t_i) + I(A_i | theta_i, t_i)
        transitions = alignments[i].getITransitionVec();
        dirCostI = dir.getIndelDirichlet(inferredN).I_theta_plus_I_A(alignments[i], transitions);

        //cout << i << " " << inferredN << " " << algnCost << " " << dirCost << endl;
        #ifdef DEBUG
        cout << "-------------------------------------------------------------------------" << endl;
        cout << "i                       = " << i << endl;
        cout << "Inferred Time           = " << inferredN << endl;
        cout << "Cost of alignment       = " << algnCost << endl;
        cout << "Dirichlet Cost (Match)  = " << dirCostM << endl;
        cout << "Dirichlet Cost (Insert) = " << dirCostI << endl;
        #endif
        objectiveCost += algnCost + dirCostM + dirCostI;
        if(objectiveCost >= numeric_limits<double>::max())
            isValid = 0;
    }

    return make_pair(objectiveCost, isValid);
}

vector<vector<double>> EMMat::perturbMatrix(const vector<vector<double>> currentMat, vector<double>& rouletteDist){
    vector<vector<double>> perturbedMat = currentMat;

    int colToPerturb = roulette_random(rouletteDist);
    assert(colToPerturb >= 0 && colToPerturb < ssealph.size());

    double sum = 0;
    for(int i = 0; i < perturbedMat.size(); i++){
        assert(colToPerturb < perturbedMat[i].size());

        perturbedMat[i][colToPerturb] = 
        gamma_random(currentMat[i][colToPerturb] * kappa, 1.0);
        assert(perturbedMat[i][colToPerturb] >= 0);
        sum += perturbedMat[i][colToPerturb];
    }

    for(int i = 0; i < perturbedMat.size(); i++){
        perturbedMat[i][colToPerturb] /= sum;
        assert(perturbedMat[i][colToPerturb] >= 0.0000);
    }

    return perturbedMat;
}

void EMMat::anneal(){

    double acceptedCost, startingCost = 0;
    
    assert(initmat.isValid());
    pair<double, int> objfuncres = cost_objective_function(initmat);
    startingCost = objfuncres.first;
    assert(objfuncres.second > 0);
   

    string tempParams = 
        "\n========================================= Parameters =========================================\n"
        "Temperature                          = " + num2str(kelvin, 0, 10) + "\n"
        "Cooling temperature Probability      = " + num2str(kelvProb, 2, 10) + "\n"
        "Stopping temperature                 = " + num2str(stoppingKelv, 4, 10) + "\n"
        "Number of Perturbations              = " + num2str(perturbations, 0, 10) + "\n"
        "Starting cost                        = " + num2str(nits2bits(startingCost), 4, 15) + " " + num2str(startingCost, 4, 15) + "\n"
        "Kappa                                = " + num2str(kappa, 4, 10) + "\n"
        "Kappa update Factor                  = " + num2str(kappaFactor, 4, 10) + "\n"
        "Tolerance                            = " + num2str(tolerance, 4, 10) + "\n"
        "Dataset size                         = " + num2str(alignments.size(), 0, 10) + "\n"
        "==============================================================================================\n";
    cout << tempParams;
    
    vector<vector<double>> matToPerturb;
    int nIgnored = 0, j = 0;
    acceptedCost = startingCost;
    acceptedMat = initmat;

    vector<double> rouletteDist = acceptedMat.getStationaryDistribution();

    string criteria;
    string iterLegend;
    string dashedline = string(94, '-');
    

    while(kelvin >= stoppingKelv){

        nIgnored = 0;
        for(j = iter; j < perturbations; j++){  

            matToPerturb = perturbMatrix(acceptedMat.getBaseMat(), rouletteDist);
            Submat perturbedMat(matToPerturb, ssealph, timeUB);

            if(!perturbedMat.isValid()){
                nIgnored++;
                continue;
            }

            pair<double, int> objfuncres = cost_objective_function(perturbedMat);
            if(objfuncres.second == 0){
                nIgnored++;
                //if(nIgnored > 1000) 
                j--;
                continue;
            }
            double newCost = objfuncres.first;
            double delta_obj = newCost - acceptedCost;

            iterLegend = "Accepted Matrix  : \n" + acceptedMat.printBaseMatrix() + "\n"; 
            iterLegend += "Perturbed Matrix : \n" + perturbedMat.printBaseMatrix() + "\n"; 

            iterLegend += 
                num2str(kelvin, 0, 8) + "\t" + to_string(j) + "\t" + num2str(nits2bits(newCost), 2, 14) + "\t" + 
                num2str(nits2bits(acceptedCost), 2, 14) + "\t" + num2str(nits2bits(-delta_obj), 2, 10) + "\t";

            if(delta_obj > -tolerance && delta_obj < tolerance) delta_obj = 0;     
            
            if(delta_obj < 0){
                acceptedCost = newCost;
                acceptedMat = perturbedMat;
                rouletteDist = acceptedMat.getStationaryDistribution();

                criteria = "Accepted\n";

            } else {
                double prob = exp(-delta_obj/kelvin);

                double c = ((double) rand() / (RAND_MAX));
                assert(prob >= 0 && prob <= 1);
                assert(c >= 0 && c <= 1);
                
                if(prob > c){
                    acceptedCost = newCost;
                    acceptedMat = perturbedMat;
                    rouletteDist = acceptedMat.getStationaryDistribution();

                    criteria = "Accepted with prob. " + num2str(prob, 5, 6) + "\n";
                
                } else {
                    criteria = "Rejected\n";
                }
            }

            cout << iterLegend << criteria << dashedline << endl;

        }
        cout << getLegend(acceptedCost, startingCost, j, acceptedMat) << endl;

        kelvin = kelvin * kelvProb;
        kappa = kappa * kappaFactor;
        iter = 0;
        if(kelvin <= 10) perturbations = 20000;
    }
}

string EMMat::getLegend(double cost, double startingCost, int iter, Submat& currmat){
    string legend =  
        "========================================= SA Legend ==========================================\n"
        "Current cost                          : " + num2str(nits2bits(cost), 9, 10) + " bits\n" +
        "Initial cost                          : " + num2str(nits2bits(startingCost), 9, 10) + " bits\n" +
        "Current temperature                   : " + num2str(kelvin, 10, 10) + "\n" +
        "Current iteration                     : " + num2str(iter, 5, 10) + "\n" +
        "----------------------------------------------------------------------------------------------\n" +
        "Accepted base matrix : \n" +
        currmat.printBaseMatrix() + "\n"
        "==============================================================================================\n";
      
    return legend;    
}

// Generates a random number from Gamma distribution with parameters alpha and beta
double EMMat::gamma_random(double alpha, double beta){
  	gamma_distribution<double> gamma_dist(alpha, beta);
	return gamma_dist(random_number_generator);
}

int EMMat::roulette_random(vector<double> alphas){
    assert(alphas.size() > 0);
    discrete_distribution<int> dist(alphas.begin(), alphas.end());
    int state = dist(random_number_generator);
    assert(state >= 0 && state < alphas.size());

    return state;
}

void EMMat::writeOutput(string& outfldr){
    string outpath = outfldr + "/SSTSUM_D_double_precision_" + currentDateTime() + ".txt";
    ofstream out(outpath, ios::out);
    assert(out);
    out << acceptedMat.printMatrix("D") << endl;
    out.close();

    outpath = outfldr + "/SSTSUM_S_double_precision_" + currentDateTime() + ".txt";
    out = ofstream(outpath,ios::out);
    assert(out);
    out << acceptedMat.printMatrix("S") << endl;
    out.close();

    outpath = outfldr + "/SSTSUM_Sinv_double_precision_" + currentDateTime() + ".txt";
    out = ofstream(outpath,ios::out);
    assert(out);
    out << acceptedMat.printMatrix("inverseS") << endl;
    out.close();

    outpath = outfldr + "/SSTSUM_stationary_double_precision_" + currentDateTime() + ".txt";
    out = ofstream(outpath,ios::out);
    assert(out);
    out << acceptedMat.printMatrix("stationary") << endl;
    out.close();
}
