#include "sseMat.h"

Submat::Submat() {
    NMATRICES = 0;
}

Submat::Submat(const string baseMatFile, int timeUB){
    NMATRICES = timeUB;
    mat = readBaseMatrix(baseMatFile);
    init();
}

Submat::Submat(vector<vector<double>>& baseMat, AAAlphabet& alph, int timeUB){

    alphabet = alph;
    NMATRICES = timeUB;
    assert(baseMat.size() > 0 && baseMat.size() == alphabet.size());

    normalize(baseMat);
    mat = baseMat;
    init();
}

Submat::Submat(AAAlphabet& alph, int timeUB){

    string    S_filepath = "../SSMMLigner/data/matrix" + to_string(alph.size()) + "/SSSUM_S_double_precision.txt";
    string    D_filepath = "./.SSMMLigner/data/matrix" + to_string(alph.size()) + "/SSSUM_D_double_precision.txt";
    string invS_filepath = "../SSMMLigner/data/matrix" + to_string(alph.size()) + "/SSSUM_Sinv_double_precision.txt";

    alphabet = alph;
    NMATRICES = timeUB;

    S        = readMatrix(S_filepath);
    D        = readMatrix(D_filepath);
    invS     = readMatrix(invS_filepath);

    generateStationaryDistribution();
    generateConditionalMatrices();
    generateJointMatrices();
    mat = matSeries[1];

    scaleMatrix();

    matSeries.erase(matSeries.begin(), matSeries.end());
    jointMSeries.erase(jointMSeries.begin(), jointMSeries.end());

    generateConditionalMatrices();
    generateJointMatrices();
    mat = matSeries[1];

    //cout << "Base = " << endl << printMatrix("base") << endl;

    assert(S.size() == alphabet.size());
    assert(D.size() == alphabet.size());
    assert(invS.size() == alphabet.size());

}

Submat::Submat(vector<vector<double>>& baseMat, vector<vector<double>>& S, vector<vector<double>>& D, vector<vector<double>>& invS, vector<double>& stationary, AAAlphabet& alph, int timeUB){
    this->S = S;
    this->D = D;
    this->invS = invS;
    this->stationarydist = stationary;

    alphabet = alph;
    NMATRICES = timeUB;

    mat = exponentiate_matrix_efficiently(1);
    assert(baseMat.size() > 0 && baseMat.size() == alphabet.size());
    assert(mat.size() == baseMat.size());
    assert(mat.size() == alphabet.size());
    for (int i = 0; i < alphabet.size(); i++) {
        for(int j = 0; j < alphabet.size(); j++){
            assert(abs(mat[i][j] - baseMat[i][j]) < 1e-10);
        }
    }

    scaleMatrix();
    
    assert(S.size() == alphabet.size());
    assert(D.size() == alphabet.size());
    assert(invS.size() == alphabet.size());

    stationarydist.erase(stationarydist.begin(), stationarydist.end());
    generateStationaryDistribution();

    for (int i = 0; i < alphabet.size(); i++) {
        assert(abs(stationarydist[i] - stationary[i]) < 1e-10);
    }

    generateConditionalMatrices();
    generateJointMatrices();
    mat = matSeries[1];
}

void Submat::init(){

    if(!eigenDecomposition()) {
        valid = false;
        return;
    }

    for(int i = 0; i < stationarydist.size(); i++){
        if(stationarydist[i] <= 0){
            valid = false;
            return;
        }
    }

    #ifdef DEBUG
    printMatrix(stationarydist);
    printMatrix(S);
    printMatrix(D);
    printMatrix(invS);
    #endif

    mat = exponentiate_matrix_efficiently(1);
    scaleMatrix();

    matSeries = vector<vector<vector<double>>>(NMATRICES + 1);  
    jointMSeries = vector<vector<vector<double>>>(NMATRICES + 1); 
}

void Submat::scaleMatrix(){
    double x = (double)0.01000000000000; // standard expected change
	double x_curr = getExpectedChange();

    if(abs(getExpectedChange() - 0.01) < 1e-10)
        return;

	double scaling_factor = x/x_curr; 
    
    assert(D.size() == mat.size());
    assert(D.size() > 0);
    vector<vector<double>> prevd = D;
    for(int j = 0; j < D.size(); j++){
        if(D[j][j] < 0)
            D[j][j] = 0.0;
		this->D[j][j] = pow(D[j][j], scaling_factor); 
	}
    mat = exponentiate_matrix_efficiently(1);

    double newExpChange = getExpectedChange();

    if(!(abs(newExpChange - 0.01) < 1e-2)) {
        cout << newExpChange << endl;
    }
    
    assert(abs(newExpChange - 0.01) < 1e-2);
}

double Submat::getExpectedChange(){
	double Pr_no_change = 0.0; 

    assert(stationarydist.size() == mat.size());
    assert(mat.size() > 0);
	for(int j = 0; j < mat.size(); j++){
		Pr_no_change += (this->stationarydist[j] * this->mat[j][j]); 
	}
	return ((double)1.0 - Pr_no_change); 
}

vector<vector<double>> Submat::readBaseMatrix(string fn) {
    ifstream infile(fn.c_str(),ios::in);
    assert(infile);
    char buff[10000];
    vector<string> inpalph;
    vector<vector<double>> tempprobs;

    size_t nRows = 0;
    size_t nCols = 0;
    while (!infile.eof()) {

        infile.getline(buff,10000);
        stringstream ss;
       
        if(buff[0] == '#') {
            ss << buff;

            char colResidue;
            ss.ignore(10, ' ');
            while(ss >> colResidue)
                inpalph.push_back(string(1, colResidue));

            alphabet = AAAlphabet(inpalph);   
            tempprobs = vector<vector<double>>(alphabet.size(), 
                  vector<double>(alphabet.size()));
            continue;
        }

        ss << buff;
        double p;
        nCols = 0;

        while(ss >> p){ 
            assert(nRows >= 0 && nRows < tempprobs.size());
            assert(nCols >= 0 && nCols < tempprobs[nRows].size());
            tempprobs[nRows][nCols++] = p;
        }
        nRows++;
    }

    normalize(tempprobs);
    
    infile.close();
    return tempprobs;
}

vector<vector<double>> Submat::readMatrix(string fn) {
    ifstream infile(fn.c_str(),ios::in);
    assert(infile);
    vector<vector<double> > mat;
    size_t nRows = alphabet.size();
    size_t nCols = nRows;
    for (size_t i = 0; i < nRows; i++) {
        vector<double> row;
        double tmp;
        for (size_t j = 0; j < nCols; j++) {
            infile >> tmp;
            row.push_back(tmp);
        }
        mat.push_back(row);
    }
    infile.close();
    return mat;
}

void Submat::normalize(vector<vector<double>>& M) {
    /* assumes input is a conditional probability matrix, and normalizes
     * by ensuring all cols add to 1 up to double precision
     * */
    int size = M.size();
    assert(size == alphabet.size());

    //find col sums
    vector<double> colsum(alphabet.size(), 0);
    for (int j = 0;  j < size; j++) {
  
        for (int i = 0;  i < size; i++) {
            if(M[i][j] < 0){
                assert(M[i][j] > -5);
                // correcting negative 0s if any
                M[i][j] = 0.0; 
            }
            //assert(M[i][j] >=  0);
            colsum[j] += M[i][j];
        }
    }

    //normalize
    for (int i = 0;  i < size; i++) {
        for (int j = 0;  j < size; j++) {
            M[i][j] /= colsum[j];
        }
    }
}

void Submat::normalize(vector<double>& M) {

    int size = M.size();
    assert(size == alphabet.size());

    double tmpsum = 0; 
    for (int j = 0;  j < size; j++) {
        if(M[j] < 0.0){
            // correcting negative 0s if any
            M[j] = (double)(-1.0)*(M[j]); 
        }
        tmpsum += M[j];
    }

    //normalize
    for (int j = 0;  j < size; j++) {
        M[j] /= tmpsum;
    }
}

bool Submat::eigenDecomposition(){

    int dim = mat.size(); 
    assert(dim == alphabet.size() && alphabet.size() > 0);

	double data[dim * dim]; 
	int c =0; 	
	for(int i = 0; i < dim; i++){
		for(int j = 0; j < dim; j++){
			data[c] = mat[i][j];
			c++;
		}
	}

	gsl_matrix_view m = gsl_matrix_view_array (data, dim, dim);
  	gsl_vector_complex *eval = gsl_vector_complex_alloc (dim);
  	gsl_matrix_complex *evec = gsl_matrix_complex_alloc (dim, dim);
  	gsl_eigen_nonsymmv_workspace * w = gsl_eigen_nonsymmv_alloc (dim);
  	int ret = gsl_eigen_nonsymmv (&m.matrix, eval, evec, w);
    assert(ret >= 0);
	gsl_eigen_nonsymmv_free (w);	

	gsl_eigen_nonsymmv_sort (eval, evec, GSL_EIGEN_SORT_ABS_DESC);  
	
	initMatrix(S, dim); 
	initMatrix(D, dim); 
	initMatrix(invS, dim); 
    
	double max_eigen_value = 0;  
	int max_eigen_value_index = 0; 	
	int min_eigen_value_index =0;
	double min_eigen_value = 100; 
		

  	for (int i = 0; i < dim; i++){

        gsl_complex eval_i = gsl_vector_complex_get (eval, i);
        gsl_vector_complex_view evec_i = gsl_matrix_complex_column (evec, i);
		double eigen_value = GSL_REAL(eval_i); 	
	
		if(max_eigen_value < eigen_value){ //take eigen value 1 for stationary dist
			max_eigen_value = eigen_value; 
			max_eigen_value_index = i; 
		}
		
        for (int j = 0; j < dim; ++j){
            if(i == j)
                D[i][j] = eigen_value;
            
            gsl_complex z = gsl_vector_complex_get(&evec_i.vector, j);
            double eigen_vec_value = GSL_REAL(z); 

            if(GSL_IMAG(z) != 0) // Complex eigen value encountered
                return false;  
            
            S[j][i] = eigen_vec_value;  
        }
    }

	invS = get_matrix_inverse(S); 	

    for(int i = 0; i < dim; i++){
        stationarydist.push_back(S[i][max_eigen_value_index]); 
    }
	
	normalize(stationarydist); 
 	gsl_vector_complex_free(eval);
  	gsl_matrix_complex_free(evec);
    return true;
}

vector<vector<double>> Submat::get_matrix_inverse(vector<vector<double>>& matrix){

	int dim = matrix.size(); 
	gsl_matrix *mat = gsl_matrix_alloc(dim, dim);

	for(int i = 0; i < dim; i++){
		for(int j = 0; j < dim; j++){
			gsl_matrix_set(mat, i, j, matrix[i][j]); 
		} 
	} 	

	gsl_permutation *p = gsl_permutation_alloc(dim);
    int s;
    // Compute the LU decomposition of this matrix
    gsl_linalg_LU_decomp(mat, p, &s);
    // Compute the  inverse of the LU decomposition
    gsl_matrix *inv = gsl_matrix_alloc(dim, dim);
    gsl_linalg_LU_invert(mat, p, inv);
    gsl_permutation_free(p);

	vector<vector<double>> matrix_inverse; 
    	for (int i = 0; i < dim; ++i) {
		    vector<double> row; 
        	for (int j = 0; j < dim; ++j) {
			    double x = round(gsl_matrix_get(inv,i,j) * 100000000000000.0) / 100000000000000.0;
            	row.push_back(x);
       		}
		matrix_inverse.push_back(row); 
	}
	gsl_matrix_free(mat);	
	gsl_matrix_free(inv);	

	return matrix_inverse; 	
}

void Submat::initMatrix(vector<vector<double>>& matrix, int k){
    vector<double> row(k, 0.0);

    for(int i = 0; i < k; i++){
		matrix.push_back(row);
	}
}

void Submat::printMatrix(vector<vector<double>>& matrix){
    
    for (int i = 0;  i < matrix.size(); i++) {
        //cout << setw(5) << string(1, alphabet.getAlphabet()[i]);

        for (int j = 0;  j < matrix.size(); j++) {
            cout << setw(18) << fixed << setprecision(15) << matrix[i][j] << " ";
        } cout << endl;
    } cout << endl;
}

void Submat::printMatrix(vector<double>& matrix){
    
    for (int i = 0;  i < matrix.size(); i++) {
        cout << setw(5) << string(1, alphabet.getAlphabet()[i]);
        cout << setw(6) << fixed << setprecision(4) << matrix[i] << endl;
    } cout << endl;
}

bool Submat::isValid(){
    return valid;
}

vector<vector<double>> Submat::compute_SubMat_N(int n){
	
    assert(n >= 0 && n <= NMATRICES);

	if(matSeries[n].size() == 0){ 
		matSeries[n] = exponentiate_matrix_efficiently(n);
	}

	return matSeries[n];
}

vector<vector<double>> Submat::exponentiate_matrix_efficiently(int n){


    if(matSeries.size() > 0)
    assert(n >= 0 && n < matSeries.size()); 

    vector<vector<double>> DpowN = D;
    for (size_t i = 0;  i < D.size(); i++){
        DpowN[i][i] = pow(D[i][i],n);
    }
    vector<vector<double>> mat =  matrixMultiply(S, matrixMultiply(DpowN,invS));
    //normalize mat.
    normalize(mat);

	return mat;
}

void Submat::generateJointMatrix(int n) {
    assert(n >= 0 && n <= NMATRICES); //sanity check

    if(jointMSeries[n].size() == 0){
    
        vector<vector<double>> jointMat;
        initMatrix(jointMat, alphabet.size());

        //compute joint using stationary and conditional
        for (int i = 0; i < alphabet.size(); i++) {
            for (int j = i; j < alphabet.size(); j++) {
                jointMat[i][j] = (double)(stationarydist[j] * matSeries[n][i][j] 
                                +        stationarydist[i] * matSeries[n][j][i])
                                /(double)2;
                if (i != j) 
                    jointMat[j][i] = jointMat[i][j];
            }
        }

        normalizeJoint(jointMat);
        jointMSeries[n] = jointMat;
    }
}

double Submat::getJointPr(char from, char to, int n){
    generateJointMatrix(n);

    //cout << string(1, from) << " " << string(1, to) << endl;
    int indx1 = alphabet[from];
    assert(indx1 >= 0 && indx1 < alphabet.size());
    int indx2 = alphabet[to];
    assert(indx2 >= 0 && indx2 < alphabet.size());

    assert(n >= 0 && n < jointMSeries.size());
    assert(indx1 >= 0 && indx1 < jointMSeries[n].size());
    assert(indx2 >= 0 && indx2 < jointMSeries[n][indx1].size());

    return abs(jointMSeries[n][indx1][indx2]);
}

double Submat::nullpr(char c){
    int indx1 = alphabet[c];
    assert(indx1 >= 0 && indx1 < alphabet.size());
    return stationarydist[indx1];
}

void Submat::normalizeJoint(vector<vector<double>>& M) {
    /* assumes input is a joint probability matrix, and normalizes
     * by ensuring all elements over the matrix add to 1 to double precision
     * */
    int size = alphabet.size();

    //ensure no elem is <0
    for (int i = 0;  i < size; i++) {
        for (int j = 0;  j < size; j++) {
            assert(M[i][j] >= 0);
        }
    }

    //find col sums
    double allsum = 0.0;
    for (int j = 0;  j < size; j++) {
        for (int i = 0;  i < size; i++) {
            allsum += M[i][j];
        }
    }

    //normalize
    for (int i = 0;  i < size; i++) {
        for (int j = 0;  j < size; j++) {
            M[i][j] /= allsum;
        }
    }
}

void Submat::generateStationaryDistribution() {
    //Find eigvec in S corresp. to eigvalue=1 in D.
    //Considering floating point rep, instead find the index of largest eigvalue
    //in D and then read the corresp. col vec in S, and normalize.
    double evaloneidx = 0;
    double maxval = D[0][0];
    for( int i = 1; i < alphabet.size(); i++) {
        if (D[i][i] > maxval) {
            maxval = D[i][i];
            evaloneidx = i;
        }
    }
    //read evaloneindx col in S matrix
    vector<double> evec;
    double sum = 0.0;
    for (int i =0; i < alphabet.size(); i++) {
        evec.push_back(S[i][evaloneidx]);
        sum += S[i][evaloneidx];
    }

    //normalize evec and write to stationarydist
    double normsum = 0.0;
    for (int i =0; i < alphabet.size(); i++) {
        double normval = (double)evec[i]/(double)sum;
        stationarydist.push_back(normval);
        normsum += normval;
    }

    assert(stationarydist.size() == alphabet.size());
    assert(normsum > 0.999999 && normsum < 1.000001);
    //check if all elems lie between [0,1]
    for (int i = 0; i < alphabet.size(); i++) {
        assert(stationarydist[i] >=0 && stationarydist[i] <=1);
    }
}

AAAlphabet Submat::getAlphabet(){
    return alphabet;
}

vector<vector<double>> Submat::getBaseMat(){
    return mat;
}

string Submat::printBaseMatrix(){
    stringstream ss;
    int prec = 15;
    
    for (int i = 0;  i < mat.size(); i++) {
        for (int j = 0;  j < mat.size(); j++) {
            ss << setw(prec+3) << fixed << setprecision(prec) << mat[i][j] << "\t";
        } ss << endl;
    } 

    return ss.str();
}

string Submat::printMatrix(string type){

    vector<vector<double>> temp;
    stringstream ss;

    if(type == "stationary"){
        for (int i = 0;  i < stationarydist.size(); i++) {
            ss << setw(18) << fixed << setprecision(15) << stationarydist[i] << "\t";
        } ss << endl;

        return ss.str();
    } else if(type == "base"){
        temp = mat;
    } else if(type == "S"){
        temp = S;
    } else if(type == "D"){
        temp = D;
    } else if(type == "inverseS"){
        temp = invS;
    } else {
        cerr << "Invalid type : " << type << ". Choose from {base, S, D, inverseS, stationary}" << endl;
        assert(false); 
    }
    
    assert(temp.size() == alphabet.size());

    for (int i = 0;  i < temp.size(); i++) {
        for (int j = 0;  j < temp.size(); j++) {
            ss << setw(18) << fixed << setprecision(15) << temp[i][j] << "\t";
        } ss << endl;
    } 
    

    return ss.str();
}

double Submat::compute_I_M(){
    double cost = 0;
    assert(mat.size() == alphabet.size());
   
    for(int j = 0; j < mat[0].size(); j++){

        vector<double> temp;

        for(int i = 0; i < mat.size(); i++){
            temp.push_back(mat[i][j]);   
        }

        sort(temp.begin(), temp.end());
        assert(mat.size() == temp.size());
        for(int k = 1; k < temp.size(); k++) {
            cost += -log(temp[k]);
        } 
    }

    return cost;
}

vector<double> Submat::getStationaryDistribution(){
    assert(stationarydist.size() == alphabet.size());
    return stationarydist;
}

void Submat::generateConditionalMatrices(){
    assert(matSeries.size()== 0); 

    for (int N = 0; N <= NMATRICES; N++) {
        vector<vector<double> > DpowN = D;
        for (size_t i = 0;  i < D.size(); i++){
            DpowN[i][i] = pow(D[i][i],N);
        }
        vector<vector<double> > mat =  matrixMultiply(S, matrixMultiply(DpowN,invS));
        //normalize mat.
        normalize(mat);
        matSeries.push_back(mat);
    }
    assert(matSeries.size()== NMATRICES+1); // matrices for N=0 to N=NMATRICES;
}

void Submat::generateJointMatrices() {
    assert(matSeries.size()== NMATRICES+1 && jointMSeries.size() == 0); 
    
    for (int N = 0; N <= NMATRICES; N++) {
        vector<vector<double> > tempmat;
        //alloc memory for mat
        for (int i =0; i < alphabet.size(); i++) {
            tempmat.push_back(vector<double>(alphabet.size(),0.0));
        }
        //compute joint using stationary and conditional
        for (int i = 0; i < alphabet.size(); i++) {
            for (int j = i; j < alphabet.size(); j++) {
                tempmat[i][j] = (double)(stationarydist[j] * matSeries[N][i][j] 
                             +       stationarydist[i] * matSeries[N][j][i])
                             /(double)2;
                if (i != j) tempmat[j][i] = tempmat[i][j];
            }
        }
        normalizeJoint(tempmat);
        jointMSeries.push_back(tempmat);
    }
    assert(jointMSeries.size()== NMATRICES+1); // matrices for N=0 to N=NMATRICES;
}