#include "threeStateDP.h"


double computeTotalProbOfAlignment(int n, SSEAlign& algnobj, Submat& ssemat){

    ssemat.compute_SubMat_N(n);
    double cost = 0;

    vector<string> align = algnobj.getMatchRegion();
    assert(align.size() == 2);

    map<string, int> matchCounts = algnobj.getMatchCounts();
    for (pair<string, int> element : matchCounts) {
        assert(element.first.length() == 2);
        double pr = ssemat.getJointPr(element.first[0], element.first[1], n);
        cost += -log(pr) * (double)element.second;
    }

    string insregion = algnobj.getInsertRegion();
    for(int f = 0; f < insregion.length(); f++){
        cost += -log(ssemat.nullpr(insregion[f]));
    }

    string delregion = algnobj.getDeleteRegion();
    for(int f = 0; f < delregion.length(); f++){
        cost += -log(ssemat.nullpr(delregion[f]));
    }

    return cost;    
}

//bisection-like search to find min_k
pair<double, int> searchForTotalProb(SSEAlign& algn, int& inferredN, Submat& ssemat, int timeUB){

    int min_k = 1;
    int lo = 1;
    int hi = timeUB; 

    double min_algnModelMsgLen = numeric_limits <double>::infinity();
   
    for(int K = 1; K <= timeUB; K++){
   
        double msglen = computeTotalProbOfAlignment(K, algn, ssemat);
        if (msglen < min_algnModelMsgLen) {
            min_algnModelMsgLen = msglen;
            min_k = K;
        }

    }
    inferredN = min_k;
    assert(inferredN >= 1 && inferredN <= timeUB);

    double sanCheck = computeTotalProbOfAlignment(inferredN, algn, ssemat);
  
    if(min_algnModelMsgLen >= numeric_limits<double>::max() || !(abs(sanCheck - min_algnModelMsgLen) <= 1e-10))
        return make_pair(min_algnModelMsgLen, 0);

    return make_pair(min_algnModelMsgLen, 1);
}

int inferTime(SSEAlign& algn, Submat& ssemat, int timeUB){

    double min = numeric_limits <double>::infinity();
    int min_k = 1;

    int lo = 1;
    int hi = timeUB; 

    for(int K = 1; K <= timeUB; K++){
   
        double msglen = computeTotalProbOfAlignment(K, algn, ssemat);
        if (msglen < min) {
            min = msglen;
            min_k = K;
        }

    }
    int inferredN = min_k;
    assert(inferredN >= 1 && inferredN <= timeUB);

    return inferredN;
}

double computeAlignmentCost(int n, SSEAlign& algnobj, Submat& ssemat, DirManager& dir){

    ssemat.compute_SubMat_N(n);
    double cost = 0;

    vector<string> align = algnobj.getMatchRegion();
    assert(align.size() == 2);

    map<string, int> matchCounts = algnobj.getMatchCounts();
    for (pair<string, int> element : matchCounts) {
        assert(element.first.length() == 2);
        double pr = ssemat.getJointPr(element.first[0], element.first[1], n);
        cost += -log(pr) * (double)element.second;
    }

    //match - I(A_i | theta_i, t_i)
    vector<int> transitions = algnobj.getMTransitionVec();
    cost += dir.getMatchDirichlet(n).I_A(algnobj, transitions);

    // indel - I(A_i | theta_i, t_i)
    transitions = algnobj.getITransitionVec();
    cost += dir.getIndelDirichlet(n).I_A(algnobj, transitions);


    string insregion = algnobj.getInsertRegion();
    double sanCost = 0;
    for(int f = 0; f < insregion.length(); f++){
        sanCost += -log(ssemat.nullpr(insregion[f]));
    }

    string delregion = algnobj.getDeleteRegion();
    for(int f = 0; f < delregion.length(); f++){
        sanCost += -log(ssemat.nullpr(delregion[f]));
    }

    double indelCost = 0;
    map<char, int> indelCounts = algnobj.getIndelCounts();
    for (pair<char, int> element : indelCounts) {
        double pr = ssemat.nullpr(element.first);
        indelCost += -log(pr) * (double)element.second;
    }

    cost += indelCost;

    assert(abs(sanCost - indelCost) < 1e-7);

    return cost;    
}

vector<string> readAFastaFile(string fname, AAAlphabet& ssealph) {
    vector<string> seqs;

    ifstream infile(fname,ios::in);
    assert(infile);
    string astr;
    string hdr = "";
    char buff[10000];
    size_t cntr = 0 ;
    while (!infile.eof()) {
    infile.getline(buff,10000);

    if (infile.eof()) {
        seqs.push_back(astr);
            seqs.push_back(hdr);
            break;
        }
        if (buff[0] == '>')  {
            string temp = buff;
            if(hdr.length() == 0)
                hdr += temp;
            else
                hdr += "_" + temp.substr(2);    

            if (cntr > 0) {
                seqs.push_back(astr);
            }
            cntr++;
            astr = "";
        }
        else {
            astr.append(buff);
        }
    }
    infile.close();
  
    return seqs;
}

vector<vector<SSEAlign>> parseBinnedAlignmentData(const string& domainTimeFilename, AAAlphabet& ssealph, int timeUB){

    ifstream infile(domainTimeFilename,ios::in);
    assert(infile);
    vector<vector<SSEAlign>> alignments(timeUB+1, vector<SSEAlign>());

    int sstCount = 0;
    char buff[10000];

    if (infile.is_open()){   //checking whether the file is open
        string str, alignpath;
        int inftime;
      
        while (getline(infile, str)) {
            stringstream ss(str);
       
            while(ss >> alignpath >> inftime){

                assert(inftime < alignments.size());
          
                ifstream infile(alignpath,ios::in);
                if(!infile) continue;
                vector<string> seqs = readAFastaFile(alignpath, ssealph);
                assert(seqs.size() == 3);

                SSEAlign algnobj(seqs[2], ssealph(seqs[0], true), ssealph(seqs[1], true));
                alignments[inftime].push_back(algnobj);
                
                sstCount++;
                
            }
        }

        infile.close(); //close the file object.
    } else{ 
        cerr << "SSTSUM: I/O Error - Error reading '" << domainTimeFilename
            << "' file." << endl;
        exit(1);
    }

    return alignments;
}

vector<SSEAlign> parseAlignmentData(const string& sstpathFilename, AAAlphabet& ssealph){

    ifstream infile(sstpathFilename,ios::in);
    assert(infile);
    vector<SSEAlign> alignments;

    int sstCount = 0;
   
    if (infile.is_open()){   //checking whether the file is open
        string alignpath, str;
        int inftime;
      
        while (getline(infile, str)) {
            stringstream ss(str);
       
            while(ss >> alignpath >> inftime){
                ifstream infile(alignpath,ios::in);
                if(!infile) {
                    cerr << "SSTSUM: I/O Error - Error reading '" << alignpath
                        << "' file." << endl;
                    exit(1);
                }
                vector<string> seqs = readAFastaFile(alignpath, ssealph);
                assert(seqs.size() == 3);

                SSEAlign algnobj(seqs[2], ssealph(seqs[0], true), ssealph(seqs[1], true));
                alignments.push_back(algnobj);
                
                sstCount++;
            }
        }

        assert(alignments.size() > 0);
        assert(alignments.size() == sstCount);
           
        infile.close(); //close the file object.
    } else{ 
        cerr << "SSSUM: I/O Error - Error reading '" << sstpathFilename
            << "' file." << endl;
        exit(1);
    }

    return alignments;
}
