#include "util.h"

int catalan(int n) {
    assert(n >= 0);
    if (n < 2) 
        return 1;

    double mult = 1;
    for (int i = 2; i <= n; i++) {
        mult *= (double)(n+i)/i;
    }
    round(mult);
    return (int)mult;
}

double encodeWTBits(int n) {
    n++;                //map from 0 onwards
    assert(n>0);        // only non-negative integers

    if (n==1) return 1;

    int cumul = 0;
    int k = 0;
    while (cumul < n) {
        cumul += catalan(k++);
    }
    return 2*(k-1)+1;
}

double encodeWTNits(int n) {
    return bits2nits(encodeWTBits(n));
}

double encodeNats(int n){
    int count, enc = 0;
    assert(n > 0);
    if(n==1) return (double) 1/log2(exp((double) 1));
    do{
        count = 0;
        while(n){
            count++;
            n = n >> 1;
        }
        enc += count; 
        n = count - 1;
    } while(n != 1);
    return (double)(enc + (double)1)/log2(exp((double) 1));
}

double encodeLogStarBits(size_t n) { 
	assert(n>=0) ;
    if (n == 0) return 0;

	double norm_const = log(2.865)/log(2); //in bits
	if (n == 1) return norm_const;

	double val = log(n)/log(2); //in bits
	double prev = val;

	while(val >= 0) {
		double t = log(prev)/log(2); //in bits
		if( t <= 0 ) break;
		val += t;
		prev = t;
	}	
	return val+norm_const;
}

double encodeLogStarNits(size_t n) { 
	assert(n>=0) ;
    if (n == 0) return 0;
    double val = encodeLogStarBits(n);
    return bits2nits(val); // in nits
}

double nits2bits(double x) {
    static double log2 = log(2) ; // log 2 to the base e
    return x/log2 ;
}

double bits2nits(double x) {
    static double log2 = log(2) ; // log 2 to the base e
    return x*log2 ;
}

string num2str(const double num, unsigned int prec, unsigned int width){
    stringstream ss;
    string ret = "";

    ss << setw(width) << fixed << setprecision(prec) << num;
    ret = ss.str();
    assert(ret.length() > 0);
    return ret;
};


vector<vector<double> > matrixMultiply(
 vector<vector<double> > A, 
 vector<vector<double> > B
) {
    size_t nRows_A = A.size();
    size_t nCols_A = A.size() > 0 ? A[0].size(): 0;

    size_t nRows_B = B.size();
    size_t nCols_B = B.size() > 0 ? B[0].size(): 0;
    assert(nCols_A == nRows_B);

    vector<vector<double> > prod;
    for (size_t i = 0 ; i < nRows_A; i++) {
        prod.push_back(vector<double>(nCols_B,0));
    }
    size_t nRows_prod = prod.size();
    size_t nCols_prod = prod.size() > 0 ? prod[0].size(): 0;
    assert(nRows_prod == nRows_A);
    assert(nCols_prod == nCols_B);

    for (size_t i = 0; i < nRows_A; i++) {
        for (size_t j = 0; j < nCols_B; j++) {
            for (size_t k = 0; k < nCols_A; k++) {
                prod[i][j] += (A[i][k]*B[k][j]);
            }
        }
    }
    return prod;
}

vector<vector<double> > matrixMultiply(
 vector<vector<double> > A, 
 int lambda
) {
    size_t nRows_A = A.size();
    size_t nCols_A = A.size() > 0 ? A[0].size(): 0;

    vector<vector<double> > prod;
    for (size_t i = 0 ; i < nRows_A; i++) {
        prod.push_back(vector<double>(nCols_A,0));
    }
    size_t nRows_prod = prod.size();
    size_t nCols_prod = prod.size() > 0 ? prod[0].size(): 0;
    assert(nRows_prod == nRows_A);
    assert(nCols_prod == nCols_A);

    for (size_t i = 0; i < nRows_A; i++) {
        for (size_t j = 0; j < nCols_A; j++) {
                prod[i][j] = lambda*A[i][j];
        }
    }
    return prod;
}


void printFSAstr2Algn(const string &fsastr, const string &S, const string &T) {
    size_t i = 0,j = 0;
    vector<string> algn(2);
    for (size_t l = 0; l < fsastr.length(); l++){
        switch (fsastr[l]) {
            case 'm': algn[0].push_back(S[i++]);
                      algn[1].push_back(T[j++]);
                      break;
            case 'i': algn[0].push_back('-');
                      algn[1].push_back(T[j++]);
                      break;
            case 'd': algn[0].push_back(S[i++]);
                      algn[1].push_back('-');
                      break;
        }
    }
    assert(S.length()==i);
    assert(T.length()==j);
    assert(algn[0].length() == fsastr.length());
    assert(algn[1].length() == fsastr.length());

    size_t nCharsPerLine = 60;
    size_t nLines        = (size_t)fsastr.length()/nCharsPerLine+1;
    size_t nc = nCharsPerLine;
    int nEquiv = 0;
    for (size_t n = 0; n < nLines; n++) {
        if (n == nLines-1) nc = algn[0].length()%nCharsPerLine;
        cout << endl;
        for (size_t c = 0; c < nc; c++) {
            if((c+1)%10 == 0) cout << ".";
            else cout << " ";
        }
        cout << endl;
        for (size_t c = 0; c < nc; c++) {
            cout << algn[0][n*nCharsPerLine+c];
        }
        cout << endl;
        for (size_t c = 0; c < nc; c++) {
            cout << algn[1][n*nCharsPerLine+c];
        }
        cout << endl;
        for (size_t c = 0; c < nc; c++) {
            if (algn[0][n*nCharsPerLine+c] == algn[1][n*nCharsPerLine+c]) {
                cout << "*";
                nEquiv++;
            }
            else {
                cout << " ";
            }
        }
    }
    cout << endl;
    cout << "nEquivalences = " << nEquiv << endl;
    cout << "\%-Identity (as a function of |A|)  = " << double(nEquiv*100)/fsastr.length() << endl;
    cout << "\%-Identity (as a function of |S|)  = " << double(nEquiv*100)/S.length() << endl;
    cout << "\%-Identity (as a function of |T|)  = " << double(nEquiv*100)/T.length() << endl;
    cout << endl;
}

vector<string> fsastr2Algn(const string &fsastr, const string &S, const string &T) {
    vector<string> algn(2);
    size_t i = 0,j = 0;
    for (size_t l = 0; l < fsastr.length(); l++){
        switch (fsastr[l]) {
            case 'm': algn[0].push_back(S[i++]);
                      algn[1].push_back(T[j++]);
                      break;
            case 'i': algn[0].push_back('-');
                      algn[1].push_back(T[j++]);
                      break;
            case 'd': algn[0].push_back(S[i++]);
                      algn[1].push_back('-');
                      break;
        }
    }
    assert(S.length()==i);
    assert(T.length()==j);
    assert(algn[0].length() == fsastr.length());
    assert(algn[1].length() == fsastr.length());

    return algn;
}

string algn2Fsastr(const vector<string> &algn) {
    string fsastr = "";
    assert(algn[0].length() == algn[1].length());

    for (size_t l = 0; l < algn[0].length(); l++) {
        if (algn[0][l] != '-' && algn[1][l] != '-') fsastr += 'm';
        else if (algn[0][l] == '-') fsastr += 'i';
        else if (algn[1][l] == '-') fsastr += 'd';
    }
    assert(algn[0].length() == fsastr.length());
    assert(algn[1].length() == fsastr.length());

    return fsastr;
}

string reverse(string r) {
    for (int i = 0, j = r.length()-1; i < j; i++, j--) {
        char tmp = r[i];
        r[i] = r[j];
        r[j] = tmp;
    }
    return r;
}

int alph2aa(char key){
    int indx = key - 'A';
    if(indx < 0) return -1; 
    assert(indx >= 0 && indx < 26);
    assert(alphMap[indx] < 20 && alphMap[indx] >= 0);
    return alphMap[indx];
}

string removeNonStandardAA(string s) {
    string sanitized_s;
    for (size_t i = 0; i < s.length(); i++) {
        switch (s[i]) {
            case 'A':
            case 'C':
            case 'D':
            case 'E':
            case 'F':
            case 'G':
            case 'H':
            case 'I':
            case 'K':
            case 'L':
            case 'M':
            case 'N':
            case 'P':
            case 'Q':
            case 'R':
            case 'S':
            case 'T':
            case 'V':
            case 'W':
            case 'Y': sanitized_s.push_back(s[i]);
                      break;
            default: //cout << "ignoring " << s[i] << " at pos = " << i+1 << endl;
                     break;
        }
    }
    return sanitized_s;
}


vector<string> file_read(const string& f_name){
    vector<string> content;
    ifstream file(f_name);
    if (file.is_open()){   //checking whether the file is open
        string str;
        while (getline(file, str)) {
            content.push_back(str);
        }
        file.close(); //close the file object.
    } else{ 
        cerr << "Compass: I/O Error - Error reading '"<<f_name<<"' file." << endl;
        exit(1);
    }
    return content;
}

// Get current date/time, format is YYYY-MM-DD.HH:mm:ss
const string currentDateTime() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%Y-%m-%d_%H-%M", &tstruct);

    return buf;
}

// Get current date/time, format is YYYY-MM-DD.HH:mm:ss
const string currentDate() {
    time_t     now = time(0);
    struct tm  tstruct;
    char       buf[80];
    tstruct = *localtime(&now);
    strftime(buf, sizeof(buf), "%Y-%m-%d", &tstruct);

    return buf;
}

string getFilenameFromPath(string fn) {
    size_t fnpos = fn.find_last_of("/\\");
    size_t extnpos = fn.find_last_of(".");
    if(fnpos!= string::npos && extnpos < fnpos) extnpos = string::npos;

    if (fnpos == string::npos && extnpos == string::npos) 
        return fn;
    else if (fnpos == string::npos)   
        return fn.substr(0,extnpos);
    else if (extnpos == string::npos) 
        return fn.substr(fnpos+1);
    else    
        return fn.substr(fnpos+1,extnpos-fnpos-1);
}

double polygamma(int degree, double value){
	return  gsl_sf_psi_n (degree,value);
}

double lattice_constant(int k){
    switch(k){
        case 1: return (1.0/12.0);
        case 2: return (5.0/(36.0*sqrt(3.0)));
        case 3: return (19.0/(192.0*pow(2.0,(1.0/3.0))));
    }
    assert(false);
    return 0;
}

void file_append(const string& f_name, const string& content){
    std::ofstream outfile;
    outfile.open(f_name, std::ios_base::app); // append instead of overwrite
    outfile << content;
}
